import nampy
from nampy import Constant, Variable
import numpy as np
from scipy.optimize.optimize import approx_fprime

"""
Implementation of function objects.
Function objects encapsulate the behaviour of an objective function that we optimize.
Simply put, implement evaluate(w, X, y) to get the numerical values corresponding to:
f, the function value (scalar) and
g, the gradient (vector).

Function objects are used with optimizers to navigate the parameter space and
to find the optimal parameters (vector). See optimizers.py.
"""

class FunObj:
    """
    Function object for encapsulating evaluations of functions and gradients
    """

    def evaluate(self, w, X, y):
        """
        Evaluates the function AND its gradient w.r.t. w.
        Returns the numerical values based on the input.
        IMPORTANT: w is assumed to be a 1d-array, hence shaping will have to be handled.
        """
        raise NotImplementedError()

    def check_correctness(self, w, X, y, epsilon=1e-6):
        n, d = X.shape
        estimated_gradient = approx_fprime(w, lambda w: self.evaluate(w, X, y)[0], epsilon=epsilon)
        _, implemented_gradient = self.evaluate(w, X, y)
        difference = estimated_gradient - implemented_gradient

        # Check the gradient
        if np.max(np.abs(difference))/np.linalg.norm(estimated_gradient) > 1e-6:
            raise Exception('User and numerical derivatives differ:\n%s\n%s' %
                (estimated_gradient[:], implemented_gradient[:]))
        else:
            print('User and numerical derivatives agree.')

class FunObjLeastSquaresMultiOutput(FunObj):
    
    def __init__(self, mapper):
        self.mapper = mapper

    def evaluate(self, w, X, Y):
        """
        Generalizes least squares error using Frobenius norm.
        Y is now an n-by-k matrix. Hence, W is a k-by-d matrix.
        """
        self.mapper.set_parameters(w)

        """YOUR CODE HERE FOR Q3.3"""
        # TODO: populate the gradients for the parameters of self.mapper.
        # This is done by composing differentiable functions to get the loss function f
        # and calling forward() and backward().
        raise NotImplementedError()

        g = self.mapper.get_gradients_flattened()

        return f, g
