
class GenericSupervisedModel:
    """
    Differentiable programming requires further separation of model behaviour components,
    thus we encapsulate prediction behaviour into a Predictor instance.

    The "model" class' responsibility becomes more singular:
    we manage retrieval and optimization of parameters at the high level.
    
    Can be used and orchestrated by a multi-layer perceptron.
    """

    def __init__(self, mapper, optimizer, fun_obj):
        self.mapper = mapper
        self.optimizer = optimizer
        self.fun_obj = fun_obj

    def predict(self, X):
        """
        predict() is the matter of getting the composed differentiable function object
        and calling forward() to get the actual numerical values based on input values.
        """
        return self.mapper.map(X).forward()

    def fit(self, X, Y): 
        """
        Refactored in A4 for less redundancy. All of the gradient-based classes
        will call self.optimize() to get the optimal parameters.
        
        Perform gradient descent using the optimizer.
        """
        n, d = X.shape
        _, k = Y.shape

        self.mapper.reset()

        # Initial guess
        w_init = self.mapper.get_parameters_flattened()
        f, g = self.fun_obj.evaluate(w_init, X, Y)

        # Reset the optimizer state and tie it to the new parameters.
        # See optimizers.py for why reset() is useful here.
        self.optimizer.reset()
        self.optimizer.set_fun_obj(self.fun_obj)
        self.optimizer.set_parameters(w_init)
        self.optimizer.set_fun_obj_args(X, Y)

        # Collect training information for debugging
        self.fs = [f]
        self.gs = [g]
        self.ws = []

        # Use gradient descent to optimize w
        for i in range(1000):
            f, g, w, break_yes = self.optimizer.step()
            self.fs.append(f)
            self.gs.append(g)
            self.ws.append(w)
            if break_yes:
                break

        self.mapper.set_parameters(w)

class GenericFactorizer:
    def __init__(self, k, mapper, optimizer_w, optimizer_z, fun_obj_w, fun_obj_z):
        self.k = k
        self.mapper = mapper
        self.optimizer_w = optimizer_w
        self.optimizer_z = optimizer_z
        self.fun_obj_w = fun_obj_w
        self.fun_obj_z = fun_obj_z

    def fit(self, X): 
        """
        Refactored in A4 for less redundancy. All of the gradient-based classes
        will call self.optimize() to get the optimal parameters.
        
        Perform gradient descent using the optimizer.
        """
        n, d = X.shape

        self.mapper.reset()

        # Initial guess
        w_init = self.mapper.get_parameters_flattened()
        f, g = self.fun_obj.evaluate(w_init, X)

        # Reset the optimizer state and tie it to the new parameters.
        # See optimizers.py for why reset() is useful here.
        self.optimizer_w.reset()
        self.optimizer_w.set_fun_obj(self.fun_obj_w)
        self.optimizer_w.set_parameters(w_init)
        self.optimizer_w.set_fun_obj_args(X)

        # Collect training information for debugging
        self.fs = [f]
        self.gs = [g]
        self.ws = []

        # Use gradient descent to optimize w
        for i in range(1000):
            f, g, w, break_yes = self.optimizer.step()
            self.fs.append(f)
            self.gs.append(g)
            self.ws.append(w)
            if break_yes:
                break


