"""
A class to separate a differentiable model's behaviour into three orthogonal categories:
1. Objective function
2. Optimization
3. Prediction via composition of differentiable operations

Mappers encapsulate the composition of differentiable operations and have two main methods:
map() -- return the differentiable variable corresponding to the "output"
reset() -- re-initialize the variables

In PyTorch, a similar concept called "Modules" are used.
"""

import nampy
import numpy as np
from nampy import DiffFunObj, Constant, Tile, Variable

class Mapper:

    def reset(self):
        """
        Re-initialize internal differentiable variables.
        """
        raise NotImplementedError()

    def map(self, X):
        """
        Outputs a composed function, which implements forward() and backward().
        """
        raise NotImplementedError()

    def set_parameters(self, w):
        """
        Takes a flattened NumPy array as an input to set the values of the internal differentiable variables.
        """
        raise NotImplementedError()

class MapperLinear(Mapper):

    def __init__(self, input_dim, output_dim, scale=1e-2):
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.scale = scale
        
    def reset(self):
        W = self.scale * np.random.randn(self.output_dim, self.input_dim)
        b = np.zeros(self.output_dim)

        # On top of numerical initializaiton, we encapsulate W and b as Variable instances
        self.W = Variable(W)
        self.b = Variable(b)

    def set_parameters(self, w): 
        """
        Take a flattened vector w and take it as weights and biases
        """
        assert w.size == self.W.value.size + self.b.value.size
        weights_flat = w[:-self.b.value.size]
        biases_flat = w[-self.b.value.size:]
        self.W.value = weights_flat.reshape(self.W.shape)
        self.b.value = biases_flat.reshape(self.b.shape)

    def get_parameters_flattened(self):
        return np.concatenate([self.W.value.reshape(-1), self.b.value.reshape(-1)])

    def get_gradients_flattened(self):
        return np.concatenate([self.W.gradient.reshape(-1), self.b.gradient.reshape(-1)])

    @property
    def size(self):
        return len(self.get_parameters_flattened())

    def map(self, X):
        """
        map() will return a composed differentiable function object,
        which can be used for forward() and backward() calls.
        """
        if not isinstance(X, DiffFunObj):
            X = Constant(X)
        n, d = X.shape
        B = Tile(self.b, n)
        Y_hat_var = (X @ self.W.T) + B
        return Y_hat_var

class MapperNeuralNet(Mapper):

    def __init__(self, encoder_mapper, predictor_mapper):
        """
        We will preserve our encoder-predictor view. Both encoder and predictor must be a Mapper instance.
        Then compose() is the matter of composing the differentiable functions for encoder and predictor.
        """
        self.encoder_mapper = encoder_mapper
        self.predictor_mapper = predictor_mapper
        
    def set_parameters(self, w): 
        """
        Take a flattened vector w and take it as weights and biases
        """
        predictor_size = self.predictor_mapper.size
        self.encoder_mapper.set_parameters(w[:-predictor_size])
        self.predictor_mapper.set_parameters(w[-predictor_size:])

    def reset(self):
        self.encoder_mapper.reset()
        self.predictor_mapper.reset()

    def map(self, X):
        Z = self.encoder_mapper.map(X)
        Y_hat = self.predictor_mapper.map(Z)
        return Y_hat

    def get_parameters_flattened(self):
        return np.concatenate([self.encoder_mapper.get_parameters_flattened(), self.predictor_mapper.get_parameters_flattened()])

    def get_gradients_flattened(self):
        return np.concatenate([self.encoder_mapper.get_gradients_flattened(), self.predictor_mapper.get_gradients_flattened()])

class MapperMultiLayer(Mapper):

    def __init__(self, input_dim, output_dim, hidden_dims, scale=1e-2):
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.hidden_dims = hidden_dims
        self.scale = scale
        
        self.layer_sizes = [input_dim, *hidden_dims, output_dim]

        # Instantiate encoder according to the specified dimensions
        self.encoder_mappers = []
        for input_dim, output_dim in zip(self.layer_sizes, self.layer_sizes[1:]):  # zip() trick to iterate consecutive pairs
            encoder_mapper = MapperLinear(input_dim, output_dim)
            self.encoder_mappers.append(encoder_mapper)

    def reset(self):
        for encoder_mapper in self.encoder_mappers:
            encoder_mapper.reset()

    def set_parameters(self, w): 
        """
        Take a flattened vector w and take it as weights and biases
        """
        encoder_sizes = [encoder_mapper.size for encoder_mapper in self.encoder_mappers]
        ws = np.array_split(w, np.cumsum(encoder_sizes))
        for w_for_encoder, encoder_mapper in zip(ws, self.encoder_mappers):
            encoder_mapper.set_parameters(w_for_encoder)

    def get_parameters_flattened(self):
        return np.concatenate([encoder_mapper.get_parameters_flattened() for encoder_mapper in self.encoder_mappers])

    def get_gradients_flattened(self):
        return np.concatenate([encoder_mapper.get_gradients_flattened() for encoder_mapper in self.encoder_mappers])

    @property
    def size(self):
        return len(self.get_parameters_flattened())

    def map(self, X):
        """
        map() will return a composed differentiable function object,
        which can be used for forward() and backward() calls.
        """
        if not isinstance(X, DiffFunObj):
            X = Constant(X)
        
        """YOUR CODE HERE FOR Q3.4"""
        # TODO: Compose differentiable operations to output the learned features after ReLU-activation.
        raise NotImplementedError()

