import itertools
import numpy as np

"""
In-house automatic differentiation module for more transparent coding exercise
Follows a similar design pattern to that of PyTorch,
a popular differentiable programming and optimization library.

Python's functional programming support is pretty limited,
so the workaround is to make a class of composable function objects.
"""

def get_indexes_from_shape(shape):
    indexes = list(itertools.product(*[np.arange(i) for i in shape]))  # for matrix input, this is [(1, 1), (1, 2), ..., (1, d), (2, 1), ..., (k, 1), ..., (k, d)]
    doubled_indexes = list((*index, *index) for index in indexes) # repeat the indexes, so we have (1, 1, 1, 1), (1, 2, 1, 2), etc. for matrix input
    return indexes, doubled_indexes

def get_generalized_identity(shape):
    return np.identity(np.multiply.reduce(shape)).reshape(shape + shape)

def get_generalized_diagonal(a):
    """
    Creates a diagonal matrix from a vector a, and creates a generalized diagonal tensor from a matrix a.
    """
    if len(a.shape) == 0:
        return a
    elif len(a.shape) == 1:
        return np.diag(a)
    else:
        r = np.zeros(a.shape + a.shape)
        np.einsum('ijij->ij', r)[:] = a
        return r

def highdiag(a):
    """
    High-dimensional generalization of np.diag()
    a -- 2D array
    output: r, a 4D array, where r[i, j] is a sparse 2D array with one non-zero entry at (i, j),
            i.e. r[i,j,i,j] = a[i,j]
    """
    r = np.zeros([*a.shape, *a.shape])
    if len(a.shape) == 0:
        return a
    elif len(a.shape) == 1:
        return np.diag(a)
    else:
        for i, j in np.ndindex(a.shape):
            r[i, j, i, j] = a[i, j]
        return r

def sum(a):
    return Sum(a)

def squared_l2_norm(a):
    return SquaredL2Norm(a)

def exp(a):
    return Exp(a)

def log(a):
    return Log(a)

def sigmoid(a):
    """YOUR CODE HERE FOR Q3.2"""
    raise NotImplementedError()

def relu(a):
    return Maximum(Constant(np.array(0.)), a)

class DiffFunObj:
    """
    A differentiable function object is different from our regular FunObj in several regards.
    First, instead of evaluate(), we will implement forward() and backward() which return
    the result of function evaluation and the gradient, respectively.

    Differentiable function objects can be composed bottom-up. That is, we instantiate
    a DiffFunObj based on another DiffFunObj function object. forward() and backward() 
    computations will be recursive.
    """
    def __init__(self):
        self.parent_fun_obj = None

    def forward(self) -> np.array or float:
        """
        forward() serves two purposes:
        1. Return the function value evaluated at the current value of the variables and constants involved
        2. Compute and remember intermediate variables including "immediate jacobian" that can be re-used during backward()
        Immediate Jacobian is my lingo for referring to the artifacts of multivariate chain rule.

        forward() is called recursively top-down.
        """
        raise NotImplementedError()

    def backward(self) -> None:
        """
        backward() computes the "immediate Jacobian" and stores it to the state of the function object.
        Immediate Jacobian is my lingo for referring to the artifacts of multivariate chain rule.

        backward() has no return value. Its main function is to recurse down to the variables involved,
        and populate their .gradient attributes with the correct arrays by using the chain rule.

        backward() must be called after forward() is called.
        backward() is also called recursively top-down.
        """
        raise NotImplementedError()

    # Overloads for native unary/binary operations
    def __add__(self, o):
        if not isinstance(o, DiffFunObj):
            o = Constant(np.asarray(o))
        return Plus(self, o)

    def __radd__(self, o):
        if not isinstance(o, DiffFunObj):
            o = Constant(np.asarray(o))
        return o + self

    def __sub__(self, o):
        if not isinstance(o, DiffFunObj):
            o = Constant(np.asarray(o))
        return self + (-o)

    def __neg__(self):
        minus_one_const = Constant(np.array(-1))
        return Multiply(minus_one_const, self)

    def __mul__(self, o):
        if not isinstance(o, DiffFunObj):
            o = Constant(np.asarray(o))
        return Multiply(self, o)

    def __rmul__(self, o):
        if not isinstance(o, DiffFunObj):
            o = Constant(np.asarray(o))
        return o * self

    def __matmul__(self, o):
        if len(self.shape) == 2 and len(o.shape) == 2:
            return MatMul(self, o)

        return Dot(self, o)

    def __pow__(self, p):
        return Power(self, p)

    def __truediv__(self, o):
        if not isinstance(o, DiffFunObj):
            o = Constant(np.asarray(o))
        return self * (o ** -1)

    def __rtruediv__(self, o):
        if not isinstance(o, DiffFunObj):
            o = Constant(np.asarray(o))
        return o * (self ** -1)

    @property
    def T(self):
        return Transpose(self)

class Constant(DiffFunObj):
    """
    For DiffFunObj composition to work seamlessly, numpy arrays must be
    wrapped within a DiffFunObj class. The constant object is a special case where
    the constructor does not take a child function object but the value inside the wrapper.
    """

    def __init__(self, value):
        """
        value is a numpy array. It can be shapeless constant, but it must implement .shape attribute.
        We will override the constructor for the constant,
        who must be evaluated as raw encapsulated avlue.
        """
        super().__init__()
        self.value = value
        self.shape = self.value.shape
        self.variable_yes = False

    def forward(self):
        """
        forward() of a constant only returns itself.
        """
        if len(self.value.shape) == 0:
            self.jacobian_me_wrt_child = np.array(0.)
        else:
            # d-dimensional vector: return d-by-d matrix
            # k-by-d matrix: return (k-by-d)-by-(k-by-d) tensor
            self.jacobian_me_wrt_child = np.zeros([*self.shape, *self.shape])

        return self.value

    def backward(self):
        """
        backward() of a constant should never be called.
        """
        raise RuntimeError("backward() of a constant should not be called")

class Variable(DiffFunObj):
    """
    The variable object is the one where the gradient is "alive".
    The constructor does not take a child function object but the value inside the wrapper.
    """

    def __init__(self, value):
        """
        We will override the constructor for the constant,
        who must be evaluated as raw encapsulated avlue.
        """
        super().__init__()
        self.value = value
        self.shape = self.value.shape
        self.gradient = 0
        self.variable_yes = True

    def forward(self):
        """
        forward() of a variable returns itself.
        """
        if len(self.value.shape) == 0:
            self.jacobian_me_wrt_child = np.array(1.)
        else:
            # https://stackoverflow.com/questions/67913457/numpy-jacobian-of-matrix-with-respect-to-self-and-2d-extension-of-diagonal-matr
            # The Jacobian of a vector identity function is the identity matrix.
            # The Jacobian of a matrix is a generalized identity tensor.
            self.jacobian_me_wrt_child = np.identity(np.multiply.reduce(self.shape)).reshape(self.shape + self.shape)
        return self.value

    def backward(self, g=None):
        """
        If backward() is called by a variable, then we found the variable with respect to which we are differenting!
        """
        if g is None:
            # In case backward() is not called by a parent.
            self.gradient = self.jacobian_me_wrt_child
        else:
            self.gradient = g

class Univariate(DiffFunObj):
    """
    Class of function objects that are differentiable.
    Each instance will maintain a state to make backward computation fast
    Univariate functions will take one argument, whether it's scalar or a numpy array
    """

    def __init__(self, child_fun_obj):
        super().__init__()
        self.child_fun_obj = child_fun_obj
        self.child_fun_obj.parent_fun_obj = self
        self.variable_yes = self.child_fun_obj.variable_yes

        # To be populated by forward()
        self.child_forwarded = None
        self.jacobian_me_wrt_child = None

    def forward(self):
        raise NotImplementedError()

    def backward(self, g=None):
        if g is None:
            jacobian_root_wrt_child = self.jacobian_me_wrt_child
        else:
            jacobian_root_wrt_me = g

            if len(jacobian_root_wrt_me.shape) <= 2 and len(self.jacobian_me_wrt_child.shape) <= 2:
                jacobian_root_wrt_child = np.dot(self.jacobian_me_wrt_child, jacobian_root_wrt_me)

            else:
                jacobian_root_wrt_child = np.tensordot(self.jacobian_me_wrt_child, jacobian_root_wrt_me)

        self.child_fun_obj.backward(jacobian_root_wrt_child)

class Transpose(Univariate):

    def __init__(self, child_fun_obj):
        super().__init__(child_fun_obj)
        self.shape = self.child_fun_obj.shape[::-1]

    def forward(self):
        self.child_forwarded = self.child_fun_obj.forward()
        if len(self.child_fun_obj.shape) == 0:
            self.jacobian_me_wrt_child = np.array(1.)
        else:
            self.jacobian_me_wrt_child = np.identity(np.multiply.reduce(self.shape)).reshape(self.shape + self.shape)
            self.jacobian_me_wrt_child = np.swapaxes(self.jacobian_me_wrt_child, 1, 0)
        return self.child_forwarded.T

class Sum(Univariate):

    def __init__(self, child_fun_obj):
        super().__init__(child_fun_obj)
        self.shape = ()  # a scalar

    def forward(self):
        """
        Evaluate child's operation, and then evaluate on it
        """
        self.child_forwarded = self.child_fun_obj.forward()

        # child_backwarded, variable_yes = self.child_fun_obj.backward()
        # if len(child_backwarded.shape) > 2:
        #     immediate_jacobian = np.ones_like(self.child_forwarded).reshape(-1).squeeze()
        # else:
        jacobian_me_wrt_child = np.ones_like(self.child_forwarded).squeeze()

        self.jacobian_me_wrt_child = jacobian_me_wrt_child

        return np.sum(self.child_forwarded)

    def backward(self, g=None):
        """
        Compute Jacobian of sum. R^d -> R, thus we get a gradient.

        If a constant is passed, then it's an identity operation
        If an array is passed, then the immediate gradient is the array itself
        """
        if self.child_forwarded is None:
            raise RuntimeError("forward() must be called before backward()")

        if g is None:
            jacobian_root_wrt_child = self.jacobian_me_wrt_child
        else:
            jacobian_root_wrt_me = g
            jacobian_root_wrt_child = np.dot(jacobian_root_wrt_me, self.jacobian_me_wrt_child)
        self.child_fun_obj.backward(jacobian_root_wrt_child)

        # General Jacobian transepose should swtich first half and second half
        # shape_idxs = np.arange(len(child_backwarded.shape))
        # first_half_idx, second_half_idx = np.split(shape_idxs, 2)
        # if len(child_backwarded.shape) > 2:
        #     child_backwarded_reshaped = child_backwarded.reshape(*child_backwarded.shape[:-2], -1).squeeze()
        # else:
        #     child_backwarded_reshaped = child_backwarded

        # child_backwarded, variable_yes = self.child_fun_obj.backward()
        # if len(child_backwarded.shape) > 2:
        #     immediate_jacobian = np.ones_like(self.child_forwarded).reshape(-1).squeeze()
        # else:
        #     immediate_jacobian = np.ones_like(self.child_forwarded).squeeze()

        # General Jacobian transepose should swtich first half and second half
        # shape_idxs = np.arange(len(child_backwarded.shape))
        # first_half_idx, second_half_idx = np.split(shape_idxs, 2)
        # if len(child_backwarded.shape) > 2:
        #     child_backwarded_reshaped = child_backwarded.reshape(*child_backwarded.shape[:-2], -1).squeeze()
        # else:
        #     child_backwarded_reshaped = child_backwarded

        # result = child_backwarded_reshaped @ immediate_jacobian

        # self.immediate_jacobian = immediate_jacobian

        # return self.finish_backward(child_backwarded_reshaped @ immediate_jacobian, variable_yes)

class Power(Univariate):

    def __init__(self, child_fun_obj, p):
        super().__init__(child_fun_obj)
        self.p = p
        self.shape = self.child_fun_obj.shape  # element-wise operation

    def forward(self):
        """
        Evaluate child's operation, and then evaluate on it
        """
        self.child_forwarded = self.child_fun_obj.forward()

        immediate_derivatives = self.p * self.child_forwarded ** (self.p - 1)
        self.jacobian_me_wrt_child = get_generalized_diagonal(immediate_derivatives)

        return self.child_forwarded ** self.p

class SquaredL2Norm(Univariate):
    """
    Dot product in general is assumed to be a bivariate operation.
    However, dot product by self is a special case where recursive Jacobian
    requires both left and right terms to return Jacobians. This is syntactically hard to capture,
    so we handle self dot product as its own operation.

    Self dot-product is the same thing as taking the squared L2 norm.
    If building a syntax override for DiffFun objects, then w @ w should be classified as L2 norm by default.    
    """

    def __init__(self, child_fun_obj):
        super().__init__(child_fun_obj)
        self.shape = ()  # a scalar

    def forward(self):
        self.child_forwarded = self.child_fun_obj.forward()
        self.jacobian_me_wrt_child = 2 * self.child_forwarded
        return self.child_forwarded @ self.child_forwarded

class Log(Univariate):
    """
    Scalar and element-wise natural logarithm.
    """

    def __init__(self, child_fun_obj):
        super().__init__(child_fun_obj)
        self.shape = self.child_fun_obj.shape  # element-wise operation

    def forward(self):
        self.child_forwarded = self.child_fun_obj.forward()
        if len(np.asarray(self.child_forwarded).reshape(-1)) == 1:
            # Scalar case
            self.jacobian_me_wrt_child = 1. / self.child_forwarded
        else:
            self.jacobian_me_wrt_child = get_generalized_diagonal(1. / self.child_forwarded).squeeze()
        return np.log(self.child_forwarded)

class Exp(Univariate):

    def __init__(self, child_fun_obj):
        super().__init__(child_fun_obj)
        self.shape = self.child_fun_obj.shape  # element-wise operation

    def forward(self):
        self.child_forwarded = self.child_fun_obj.forward()
        if len(np.asarray(self.child_forwarded).reshape(-1)) == 1:
            self.jacobian_me_wrt_child = np.exp(self.child_forwarded)
        else:
            self.jacobian_me_wrt_child = np.diag(np.exp(self.child_forwarded))
        return np.exp(self.child_forwarded)

class Tile(Univariate):

    def __init__(self, child_fun_obj, n):
        """
        Takes a variable/constant and replicates tht output in a new axis 'n' times.
        That is, given a d-by-1 vector, the output will be an n-by-d matrix.
        """
        super().__init__(child_fun_obj)
        self.n = n
        self.shape = (n, *self.child_fun_obj.shape)

    def forward(self):
        self.child_forwarded = self.child_fun_obj.forward()

        # Compute Jacobian
        self.jacobian_me_wrt_child = np.zeros(self.child_fun_obj.shape + self.shape)
        tmp = np.ones(self.shape)
        np.einsum('iji->ji', self.jacobian_me_wrt_child)[:] = tmp
        return np.broadcast_to(self.child_forwarded, self.shape)

class Bivariate(DiffFunObj):
    """
    Bivariate functions will take two DiffFunObj classes into constructor.
    """

    def __init__(self, left, right):
        super().__init__()
        self.left = left
        self.left.parent_fun_obj = self
        self.right = right
        self.right.parent_fun_obj = self
        self.variable_yes = self.left.variable_yes or self.right.variable_yes

        # to be populated by forward()
        self.left_forwarded = None  # useful for debugging to bookkeep
        self.right_forwarded = None  # useful for debugging to bookkeep
        self.jacobian_me_wrt_left = None
        self.jacobian_me_wrt_right = None

    def forward(self):
        raise NotImplementedError()

    def backward(self, g=None):
        if g is None:
            if self.left.variable_yes:
                jacobian_root_wrt_left = self.jacobian_me_wrt_left
                self.left.backward(jacobian_root_wrt_left)
            
            if self.right.variable_yes:
                jacobian_root_wrt_right = self.jacobian_me_wrt_right
                self.right.backward(jacobian_root_wrt_right)

        else:
            jacobian_root_wrt_me = g

            if self.left.variable_yes:
                if len(jacobian_root_wrt_me.shape) <= 2 and len(self.jacobian_me_wrt_left.shape) <= 2:
                    jacobian_root_wrt_left = np.dot(jacobian_root_wrt_me, self.jacobian_me_wrt_left)
                else:
                    jacobian_root_wrt_left = np.tensordot(jacobian_root_wrt_me, self.jacobian_me_wrt_left)
                self.left.backward(jacobian_root_wrt_left)

            if self.right.variable_yes:
                if len(jacobian_root_wrt_me.shape) <= 2 and len(self.jacobian_me_wrt_right.shape) <= 2:
                    jacobian_root_wrt_right = np.dot(jacobian_root_wrt_me, self.jacobian_me_wrt_right)
                else:
                    jacobian_root_wrt_right = np.tensordot(jacobian_root_wrt_me, self.jacobian_me_wrt_right)
                self.right.backward(jacobian_root_wrt_right)

class Plus(Bivariate):
    """
    Handles scalar plus scalar and vector plus vector.
    """

    def __init__(self, left, right):
        super().__init__(left, right)

        # Scalar broadcast multiplication
        if len(self.left.shape) == 0:
            self.shape = self.right.shape

        elif len(self.right.shape) == 0:
            self.shape = self.left.shape

        elif len(self.left.shape) == 1 and len(self.right.shape) == 1:
            self.shape = self.left.shape

        elif len(self.left.shape) == 2 and len(self.right.shape) == 2:
            self.shape = self.left.shape

        else:
            raise RuntimeError("+ operation has mismatching shapes between left and right: {:s}, {:s}".format(str(left.shape), str(right.shape)))

    def forward(self):
        self.left_forwarded = self.left.forward()
        self.right_forwarded = self.right.forward()
        
        if self.left.variable_yes:
            self.jacobian_me_wrt_left = get_generalized_identity(self.left_forwarded.shape)
            
        if self.right.variable_yes:
            self.jacobian_me_wrt_right = get_generalized_identity(self.right_forwarded.shape)
        # if self.left.variable_yes and right_variable_yes
        # Assume that left and right have the same shape, otherwise broadcasting won't work
        # Then the resulting vector has the same shape and left and right
        return self.left_forwarded + self.right_forwarded

    def backward(self, g=None):
        """
        Yield Jacobian for vector-vector addition. It's a pass-through.
        """
        if g is None:
            jacobian_root_wrt_me = None
        else:
            jacobian_root_wrt_me = g
        if self.left.variable_yes:
            self.left.backward(jacobian_root_wrt_me)
        if self.right.variable_yes:
            self.right.backward(jacobian_root_wrt_me)

class Dot(Bivariate):
    """
    Handles both matrix-vector product and vector-vector product.
    Matrix multiplication should be handled via MatMul.
    """

    def __init__(self, left, right):
        super().__init__(left, right)

        if len(self.left.shape) >= 2 and len(self.right.shape) >= 2:
            raise RuntimeError("DotProduct is invoked for matrix-matrix multiplication. Shapes: {:s}, {:s}".format(str(left.shape), str(right.shape)))

        # Shape check
        if self.left.shape[-1] != self.right.shape[0]:
            raise RuntimeError("DotProduct operation has mismatching shapes between left and right: {:s}, {:s}".format(str(left.shape), str(right.shape)))

        # Vector-Vector
        if len(self.left.shape) == 1 and len(self.right.shape) == 1:
            self.shape = ()  # result is a scalar

        # Matrix-Vector
        if len(self.left.shape) == 2 and len(self.right.shape) == 1:
            self.shape = (self.left.shape[0],)

        # Vector-Matrix
        if len(self.left.shape) == 1 and len(self.right.shape) == 2:
            self.shape = (self.right.shape[-1],)

    def forward(self):
        self.left_forwarded = self.left.forward()
        self.right_forwarded = self.right.forward()
        if self.left.variable_yes:
            self.jacobian_me_wrt_left = self.right_forwarded.T
        if self.right.variable_yes:
            self.jacobian_me_wrt_right = self.left_forwarded.T
        return self.left_forwarded @ self.right_forwarded

    def backward(self, g=None):
        """
        \nabla_w w^Tx_i = w
        \nabla_w w^Tw = 2 * w
        """
        if self.left.variable_yes:
            if g is None:
                jacobian_root_wrt_left = self.jacobian_me_wrt_left
            else:
                jacobian_root_wrt_me = g
                jacobian_root_wrt_left = np.dot(self.jacobian_me_wrt_left, jacobian_root_wrt_me)
            self.left.backward(jacobian_root_wrt_left)

        if self.right.variable_yes:
            if g is None:
                jacobian_root_wrt_right = self.jacobian_me_wrt_right
            else:
                jacobian_root_wrt_me = g
                jacobian_root_wrt_right = np.dot(self.jacobian_me_wrt_right, jacobian_root_wrt_me)
            self.right.backward(jacobian_root_wrt_right)

class Multiply(Bivariate):
    """
    Scalar product and vector element-wise multiplication
    """

    def __init__(self, left, right):
        super().__init__(left, right)

        # Scalar broadcast multiplication
        if len(self.left.shape) == 0:
            self.shape = self.right.shape

        elif len(self.right.shape) == 0:
            self.shape = self.left.shape

        elif len(self.left.shape) == 1 and len(self.right.shape) == 1:
            self.shape = self.left.shape

        elif len(self.left.shape) == 2 and len(self.right.shape) == 2:
            self.shape = self.left.shape

        else:
            # TODO: support vector-matrix broadcast multiplication?
            raise RuntimeError("* operation has mismatching shapes between left and right: {:s}, {:s}".format(str(left.shape), str(right.shape)))

    def forward(self):
        self.left_forwarded = self.left.forward()
        self.right_forwarded = self.right.forward()
        if self.left.variable_yes:
            self.jacobian_me_wrt_left = get_generalized_diagonal(self.right_forwarded)
        if self.right.variable_yes:
            self.jacobian_me_wrt_right = get_generalized_diagonal(self.left_forwarded)
        return self.left_forwarded * self.right_forwarded

    def backward(self, g=None):
        if g is None:
            if self.left.variable_yes:
                jacobian_root_wrt_left = self.jacobian_me_wrt_left
                self.left.backward(jacobian_root_wrt_left)
            
            if self.right.variable_yes:
                jacobian_root_wrt_right = self.jacobian_me_wrt_right
                self.right.backward(jacobian_root_wrt_right)

        else:
            jacobian_root_wrt_me = g

            if self.left.variable_yes:
                jacobian_root_wrt_left = np.dot(jacobian_root_wrt_me, self.jacobian_me_wrt_left)
                self.left.backward(jacobian_root_wrt_left)

            if self.right.variable_yes:
                jacobian_root_wrt_right = np.dot(jacobian_root_wrt_me, self.jacobian_me_wrt_right)
                self.right.backward(jacobian_root_wrt_right)

class Maximum(Bivariate):
    """
    Scalar product and vector element-wise multiplication
    """

    def __init__(self, left, right):
        super().__init__(left, right)

        # Scalar broadcast
        if len(self.left.shape) == 0:
            self.shape = self.right.shape

        elif len(self.right.shape) == 0:
            self.shape = self.left.shape

        # Vector-vector
        elif len(self.left.shape) == 1 and len(self.right.shape) == 1:
            self.shape = self.left.shape

        # Matrix-matrix
        elif len(self.left.shape) == 2 and len(self.right.shape) == 2:
            self.shape = self.left.shape

        else:
            # TODO: support vector-matrix broadcast?
            raise RuntimeError("max operation has mismatching shapes between left and right: {:s}, {:s}".format(str(left.shape), str(right.shape)))

    def forward(self):
        """YOUR CODE HERE FOR Q3.2"""
        # Hint: if you read the backward() function in the Bivariate class, you will see that
        # forward() is supposed to populate self.jacobian_me_wrt_left and self.jacobian_me_wrt_right.
        # Hence your forward() call should:
        # 1. Compute the maximum operation (take max between left and right)
        # 2. Compute the Jacobian of the output wrt left and wrt right
        raise NotImplementedError()
        

class MatMul(Bivariate):
    """
    Matrix-matrix multiplication
    """

    def __init__(self, left, right):
        super().__init__(left, right)

        if self.left.shape[-1] != self.right.shape[0]:
            raise RuntimeError("MatMul operation has mismatching shapes between left and right: {:s}, {:s}".format(str(left.shape), str(right.shape)))

        self.shape = (self.left.shape[0], self.right.shape[-1])

    def forward(self):
        """YOUR CODE HERE FOR Q3.2"""
        # Hint: you will find np.einsum() extremely useful for broadcast multiplication
        raise NotImplementedError()

        
    def backward(self, g=None):
        """YOUR CODE HERE FOR Q3.2"""
        # Hint: it is easier than you think! Write the chain rule down on the paper, then it becomes much easier to see.
        raise NotImplementedError
        if self.left.variable_yes:
            if g is None:
                jacobian_root_wrt_left = self.jacobian_me_wrt_left
            else:
                jacobian_root_wrt_me = g
                jacobian_root_wrt_left = None  # TODO: replace this
            self.left.backward(jacobian_root_wrt_left)

        if self.right.variable_yes:
            if g is None:
                jacobian_root_wrt_right = self.jacobian_me_wrt_right
            else:
                jacobian_root_wrt_me = g
                jacobian_root_wrt_right = None  # TODO: replace this
            self.right.backward(jacobian_root_wrt_right)




